'''
Emits RISCV machine code that blinks an LED.
'''

from dataclasses import dataclass

@dataclass
class Register:
    name: str
    # the 3 bits that identify the register.
    n: int

reg_zero = Register('zero', 0)
reg_t0 = Register('t0', 5)
reg_t1 = Register('t1', 6)
reg_t2 = Register('t2', 7)

@dataclass
class and_bits:
    rd: Register
    rs1: Register
    rs2: Register
    name = 'and'
    opcode = 0b0110011
    fmt = 'R'
    funct3 = 0x7
    funct7 = 0x00

@dataclass
class or_bits:
    rd: int
    rs1: Register
    rs2: Register
    name = 'or'
    opcode = 0b0110011
    fmt = 'R'
    funct3 = 0x6
    funct7 = 0x00

@dataclass
class lui:
    rd: Register
    imm: int
    name = 'lui'
    opcode = 0b0110111
    fmt = 'U'

@dataclass
class addi:
    rd: Register
    rs1: Register
    imm: int
    name = 'addi'
    opcode = 0b0010011
    fmt = 'I'
    funct3 = 0x0

# pseudo-instruction.
# can't load a 32-bit immediate in a single instruction, it has to be
# broken down into smaller ones.
@dataclass
class li:
    rd: Register
    imm: int
    def build(self):
        # only non-negative immediates supported for now.
        assert self.imm >= 0
        res = [lui(self.rd, self.imm >> 12)]
        remaining = self.imm & 0xFFF
        while remaining > 0:
            # at most 2047 since anything greater than that is interpreted as negative.
            x = min(remaining, 2047)
            res.append(addi(self.rd, self.rd, x))
            remaining -= x
        return res

@dataclass
class lw:
    rd: Register
    rs1: Register
    # can be 0
    imm: int
    name = 'lw'
    opcode = 0b0000011
    fmt = 'I'
    funct3 = 0x2

@dataclass
class sw:
    rs1: Register
    rs2: Register
    # can be 0
    imm: int
    name = 'sw'
    opcode = 0b0100011
    fmt = 'S'
    funct3 = 0x2

@dataclass
class jal:
    rd: int
    imm: int
    name = 'jal'
    opcode = 0b1101111
    fmt = 'J'


@dataclass
class bne:
    rs1: int
    rs2: int
    imm: int
    name = 'bne'
    opcode = 0b1100011
    funct3 = 0x1
    fmt = 'B'


def set_bit(memory_address: int, bit: int, enabled: bool):
    out = []
    out.extend([
        *li(reg_t0, memory_address).build(),
        lw(reg_t1, reg_t0, 0),
    ])
    if enabled:
        out.extend([
            *li(reg_t2, 1 << bit).build(),
            or_bits(reg_t2, reg_t1, reg_t2),
        ])
    else:
        out.extend([
            *li(reg_t2, 0xFFFFFFFF ^ (1 << bit)).build(),
            and_bits(reg_t2, reg_t1, reg_t2),
        ])
    out.extend([
        sw(reg_t0, reg_t2, 0),
    ])
    return out


def busy_loop(num_instructions):
    out = []
    out.extend([
        addi(reg_t0, reg_zero, 0),
        *li(reg_t1, num_instructions).build(),
        # Jump to here...
        addi(reg_t0, reg_t0, 1),
        bne(reg_t0, reg_t1, 0x1FFC),
    ])
    return out


def blinky(pin_number: int, pin_on: bool):
    assert 0 <= pin_number < 31
    out = []

    # Disable watchdog timer that will otherwise kill the program in a bit.
    # From the technical reference. "After booting, IMG_WDT_FLASHBOOT_MOD_EN
    # and LP_WDT_RWDT_FLASHBOOT_MOD_EN should be cleared to stop the flash
    # boot protection procedure for both MWDT0 and RWDT respectively."
    # This only disables the former, but it seems to work!
    TIMG0_BASE_ADDR = 0x6000_8000
    TIMG_WDTCONFIG0_REG_ADDR = TIMG0_BASE_ADDR + 0x48
    TIMG_WDT_FLASHBOOT_MOD_EN = 14
    out.extend(set_bit(TIMG_WDTCONFIG0_REG_ADDR, TIMG_WDT_FLASHBOOT_MOD_EN, False))

    IO_MUX_BASE_ADDR = 0x6009_0000
    IO_MUX_GPIOn_REG_ADDR = IO_MUX_BASE_ADDR + 0x4 + 4*pin_number
    # Disable pull up & pull down resistors, and disable input.
    FUN_WPD = 7
    FUN_WPU = 8
    FUN_IE = 9
    out.extend(set_bit(IO_MUX_GPIOn_REG_ADDR, FUN_WPD, False))
    out.extend(set_bit(IO_MUX_GPIOn_REG_ADDR, FUN_WPU, False))
    out.extend(set_bit(IO_MUX_GPIOn_REG_ADDR, FUN_IE, False))

    # Configure IO MUX function as GPIO.
    out.extend(set_bit(IO_MUX_GPIOn_REG_ADDR, 12, True))
    out.extend(set_bit(IO_MUX_GPIOn_REG_ADDR, 13, False))
    out.extend(set_bit(IO_MUX_GPIOn_REG_ADDR, 14, False))

    # Simple GPIO output
    GPIO_MATRIX_BASE_ADDR = 0x6009_1000

    # Set GPIO matrix GPIO_FUNCn_OUT_SEL with a special peripheral index 128 (0x80).
    GPIO_FUNCn_OUT_SEL_ADDR = GPIO_MATRIX_BASE_ADDR + 0x0554 + 4*pin_number
    out.extend([
        *li(reg_t0, GPIO_FUNCn_OUT_SEL_ADDR).build(),
        *li(reg_t1, 0x80).build(),
        sw(reg_t0, reg_t1, 0),
    ])

    # Enable output
    GPIO_ENABLE_REG_ADDR = GPIO_MATRIX_BASE_ADDR + 0x20
    out.extend(set_bit(GPIO_ENABLE_REG_ADDR, pin_number, True))

    GPIO_OUT_REG_ADDR = GPIO_MATRIX_BASE_ADDR + 0x4

    # Blink the LED.
    set_high_label = len(out)
    out.extend(set_bit(GPIO_OUT_REG_ADDR, pin_number, True))
    out.extend(busy_loop(40_000_000))
    out.extend(set_bit(GPIO_OUT_REG_ADDR, pin_number, False))
    out.extend(busy_loop(40_000_000))
    # Jump back to set_high_label.
    # Compute the 21-bit twos complement for the relative jump.
    # Note each instruction = 4 bytes.
    relative_jump = (0x1FFFFF ^ 4*(len(out) - set_high_label)) + 1
    out.append(jal(reg_zero, relative_jump))
    return out


def bit_range(v, start_inc, end_inc) -> int:
    """
    Selects bits in the inclusive range
    Example: 0b011001, 2, 4 => 0b110
    """
    return (v & ((1 << (end_inc + 1)) - 1)) >> start_inc


def pack_bit_ranges(*ranges, assert_num_bits=None) -> int:
    """
    Packs bits selected by bit_range side-by-side. Starts packing at bit 0
    up to assert_num_bits.
    """
    res = 0
    i = 0
    for (v, start_inc, end_inc) in ranges:
        res |= bit_range(v, start_inc, end_inc) << i
        i += end_inc - start_inc + 1
    if assert_num_bits:
        assert i == assert_num_bits
    return res


def encode_instr_as_integer(instr):
    match instr.fmt:
        case 'R':
            return pack_bit_ranges(
                (instr.opcode, 0, 6),
                (instr.rd.n, 0, 4),
                (instr.funct3, 0, 2),
                (instr.rs1.n, 0, 4),
                (instr.rs2.n, 0, 4),
                (instr.funct7, 0, 6),
                assert_num_bits=32,
            )
        case 'I':
            return pack_bit_ranges(
                (instr.opcode, 0, 6),
                (instr.rd.n, 0, 4),
                (instr.funct3, 0, 2),
                (instr.rs1.n, 0, 4),
                (instr.imm, 0, 11),
                assert_num_bits=32,
            )
        case 'S':
            return pack_bit_ranges(
                (instr.opcode, 0, 6),
                (instr.imm, 0, 4),
                (instr.funct3, 0, 2),
                (instr.rs1.n, 0, 4),
                (instr.rs2.n, 0, 4),
                (instr.imm, 5, 11),
                assert_num_bits=32,
            )
        case 'B':
            return pack_bit_ranges(
                (instr.opcode, 0, 6),
                (instr.imm, 11, 11),
                (instr.imm, 1, 4),
                (instr.funct3, 0, 2),
                (instr.rs1.n, 0, 4),
                (instr.rs2.n, 0, 4),
                (instr.imm, 5, 10),
                (instr.imm, 12, 12),
                assert_num_bits=32,
            )
        case 'U':
            return pack_bit_ranges(
                (instr.opcode, 0, 6),
                (instr.rd.n, 0, 4),
                (instr.imm, 0, 19),
                assert_num_bits=32,
            )
        case 'J':
            return pack_bit_ranges(
                (instr.opcode, 0, 6),
                (instr.rd.n, 0, 4),
                (instr.imm, 12, 19),
                (instr.imm, 11, 11),
                (instr.imm, 1, 10),
                (instr.imm, 20, 20),
                assert_num_bits=32,
            )


def encode(instructions):
    res = bytearray()
    for instr in instructions:
        v = encode_instr_as_integer(instr)
        # Bits go in little endian byte order.
        res.extend([
            (v & 0x000000ff),
            (v & 0x0000ff00) >> 8,
            (v & 0x00ff0000) >> 16,
            (v & 0xff000000) >> 24,
        ])
    return res


def asm(instructions):
    res = []
    for instr in instructions:
        match instr.fmt:
            case 'R':
                res.append(f'{instr.name} {instr.rd.name},{instr.rs1.name},{instr.rs2.name}')
            case 'I':
                if instr.name == 'lw':
                    res.append(f'{instr.name} {instr.rd.name},{instr.imm}({instr.rs1.name})')
                else:
                    res.append(f'{instr.name} {instr.rd.name},{instr.rs1.name},{instr.imm}')
            case 'S':
                res.append(f'{instr.name} {instr.rs2.name},{instr.imm}({instr.rs1.name})')
            case 'U':
                res.append(f'{instr.name} {instr.rd.name},{instr.imm}')
            case 'B':
                res.append(f'{instr.name} {instr.rs1.name},{instr.rs2.name},{instr.imm}')
            case 'J':
                res.append(f'{instr.name} {instr.rd.name},{instr.imm}')
    return '\n'.join(res) + '\n'

# Direct boot is the simplest image format supported by ESP32-C6.
# If you put two magic 0xaedb041d words in the beginning, machine
# code right after it is executed.
# https://www.espressif.com/sites/default/files/documentation/esp32-c6_technical_reference_manual_en.pdf#subsubsection.9.2.2
direct_boot_header = bytearray(2*[0x1d, 0x04, 0xdb, 0xae])


if __name__ == '__main__':
    instrs = blinky(18, True)
    with open('blink.bin', 'wb') as f:
        # machine code output
        f.write(direct_boot_header + encode(instrs))
    with open('blink.asm', 'wb') as f:
        # assembly output
        f.write(asm(instrs).encode('utf-8'))
